# Vehicle Simulator - http://lkcl.net/ev
# Copyright (C) 2011 Luke Kenneth Casson Leighton <lkcl@lkcl.net>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This application has been written to provide the proof-of-concept for the
# Ultra-efficient Hybrid Electric Vehicle:
#
#    http://lkcl.net/ev
#
# However, it is also a generic Vehicle Simulator, and can be used to
# simulate other cars.  The code can be used in three different ways:
# 1) command-line.  type "python sim.py" for an example.
# 2) GUI app.  install Pyjamas Desktop (http://pyjs.org) and run "Simulator.py"
# 3) Web app.  install Pyjamas (http://pyjs.org) and build Simulator.py

header = """General-purpose Vehicle Simulator, proof-of-concept for<br />
Ultra-efficient Hybrid EV <a href="http://lkcl.net/ev">http://lkcl.net/ev</a>.
<p>
Notes: Enter the vehicle characteristics on the left input-boxes, and the
starting conditions on the right-hand ones.  yes, 1st gear starts at zero.
If you need different gear ratios (or a different number of gears) from the
ones shown, simply enter them as a list with brackets as shown.  Hit "run"
to run the simulation, then if you need to change anything hit "clear" first.
</p>
<p>
Regarding the fuel efficiency figure: 0.28 litres per kilowatt per hour
was picked off of a wikipedia page on Diesel Generators
(http://en.wikipedia.org/wiki/Diesel_generator).  The efficiency of the
Diesel Engine itself is *already* taken into account in this figure.
Thus, the Drivetrain efficiency of 80% is a conservative figure that includes
the efficiency of the electric motor (as a static figure instead of a
varying one depending on torque and RPM), but PLEASE NOTE that above about
1500kg there will be additional drivetrain losses due to increased friction,
and you will need to take that into account by entering an alternative
estimated figure.
</p>
<p>
Also, if using a single direct-drive motor (only one gear ratio), then that
motor will definitely be operating outside of the assumptions made for this
simulator (namely, a fixed drivetrain efficiency figure).  Your mileage
may vary - the source code is always available if you wish to make improvements
(git clone http://lkcl.net/ev/vehicle_simulator/.git)
</p>
"""

import pyjd # this is dummy in pyjs

from pyjamas.ui.Button import Button
from pyjamas.ui.RootPanel import RootPanel
from pyjamas.ui.Label import Label
from pyjamas.ui.Grid import Grid
from pyjamas.ui.CellFormatter import CellFormatter
from pyjamas.ui.RowFormatter import RowFormatter
from pyjamas.ui.HTML import HTML
from pyjamas.ui.CheckBox import CheckBox
from pyjamas.ui.AbsolutePanel import AbsolutePanel
from pyjamas.ui.TextBox import TextBox
from pyjamas.ui.HorizontalPanel import HorizontalPanel
from pyjamas import Window

from sim import accelerate_to_top_speed
from physics import slope_to_gradient
from vehicle import Vehicle
from state import State

headings = ['Time<br />(s)', 'Gear<br/>', 'Req<br /> (W)', 'Pow <br />(W)', 'Accel <br />(m/s/s)', 'Speed <br />(mph)', 'Dist <br />(m)', 'Econ@const.spd <br />(mpg)']
formats = ['%4.1f', '%d', '%6.0f', '%6.0f', '%6.4f', '%6.2f', '%8.1f', '%6.1f']


class StateData(Grid):
    def __init__(self):
        Grid.__init__(self)
        self.fields = {
              'speed':0.0,
              'gear_selected':0,
              'fuel_efficiency':0.28,
              'slope_degrees': 0.0001}
        self.input_fields = {}
        self.resize(len(self.fields), 2)
        row = 0
        for fname, v in self.fields.items():
            f = fname.split('_')
            f = map(lambda x: x[0].upper() + x[1:], f)
            f = ' '.join(f)
            i = TextBox(Text=str(v))
            self.input_fields[fname] = i
            self.setText(row, 0, f)
            self.setWidget(row, 1, i)
            row += 1

    def create_state(self):
        vals = {'throttle': 1.0,
               'distance_travelled':0.0
            }
        for (fname, ib) in self.input_fields.items():
            val = ib.getText()
            if val and val[0] == '[' and val[-1] == ']':
                l = []
                for v in val[1:-1].split(','):
                    l.append(float(v.strip()))
                val = l
            else:
                if val.find(".") != -1:
                    val = float(val)
                else:
                    val = int(val)
            vals[fname] = val
        return State(**vals)

class InputData(Grid):
    def __init__(self):
        Grid.__init__(self)
        self.fields = {
                'weight': 550,
                'drivetrain_efficiency':0.8,
                'drag_coefficient':0.2,
                'surface_area':1.1,
                'rolling_resistance':0.007,
                'gear_ratios':'[4.62, 2.69, 1.53, 1.0, 0.855]',
                'axle_ratio':4.16,
                'max_rpm':3500.0,
                'max_accel':6.0, # m/sec
                'max_power':15000.0,
                'max_available_power':15000.0,
                'peak_torque':60.0, # Nm
                'wheel_radius':0.33}
        self.input_fields = {}
        self.resize(len(self.fields), 2)
        row = 0
        for fname, v in self.fields.items():
            f = fname.split('_')
            f = map(lambda x: x[0].upper() + x[1:], f)
            f = ' '.join(f)
            i = TextBox(Text=str(v))
            self.input_fields[fname] = i
            self.setText(row, 0, f)
            self.setWidget(row, 1, i)
            row += 1

    def create_vehicle(self):
        vals = {}
        for (fname, ib) in self.input_fields.items():
            val = ib.getText()
            if val and val[0] == '[' and val[-1] == ']':
                l = []
                for v in val[1:-1].split(','):
                    l.append(float(v.strip()))
                val = l
            else:
                val = float(val)
            vals[fname] = val
        return Vehicle(**vals)

class GridWidget(AbsolutePanel):

    def __init__(self):
        AbsolutePanel.__init__(self)

        self.clearb = Button("Clear", self)
        self.runb = Button("Run", self)
        self.vinput = InputData()
        self.sinput = StateData()
        
        self.g = Grid()
        self.g.resize(1, 8)
        for (c, fmt) in enumerate(headings):
            self.g.setWidget(0, c, HTML("<b>%s</b>" % fmt))
        self.g.setBorderWidth(2)
        self.g.setCellPadding(4)
        self.g.setCellSpacing(1)
        
        hp = HorizontalPanel(Spacing=5)
        hp.add(self.vinput)
        hp.add(self.sinput)

        self.notes = HTML()

        self.add(HTML(header))
        self.add(hp)
        self.add(self.clearb)
        self.add(self.runb)
        self.add(self.notes)
        self.add(self.g)

    def onClick(self, sender):
        if sender == self.clearb:
            self.g.resize(1, 8)
            return
        if sender == self.runb:
            self.run_sim()
            return
        

    def run_sim(self):
        
        time_delta = 0.1
        top_speed = 200.0

        v = self.vinput.create_vehicle()
        s = self.sinput.create_state()

        if abs(s.slope_degrees) < 0.000001:
            s.slope_degrees = 0.000001
        grad = slope_to_gradient(s)
            
        results = accelerate_to_top_speed(s, v, time_delta, top_speed)
        
        for r in results:
            rc = self.g.getRowCount()
            self.g.resizeRows(rc+1)
            for (c, fmt) in enumerate(formats):
                self.g.setText(rc, c, fmt % r[c])

        notes = "<b>Notes</b><br />"
        notes += "Fuel efficiency %.2f is in Litres / kW / hr<br />" % \
                        s.fuel_efficiency
        notes += "Gradient is %.1f%%<br />" % (grad * 100.0)
        notes += "Wheel Radius is %.1fin<br />" % (v.wheel_radius/.0254)
        if results[-1][5] != 200.0:
            notes += "Top speed reached: %.1fmph<br />" % (results[-1][5])
        if results[-1][4] < 0.01:
            notes += "Top Speed limited by engine capacity<br />"
        
        self.notes.setHTML(notes)

def AppInit():
    return GridWidget()

if __name__ == '__main__':
    pyjd.setup("./public/Simulator.html")
    g = GridWidget()
    RootPanel().add(g)
    pyjd.run()

